#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from py4j.java_gateway import JavaClass

from pygw.base import GeoWaveObject
from pygw.base.type_conversions import PrimitiveByteArrayType
from pygw.config import java_gateway

from .query_constraints import QueryConstraints

_pbat = PrimitiveByteArrayType()


class QueryConstraintsFactory(GeoWaveObject):
    """
    Base factory for constructing general query constraints to be used by the
    query builder.  Do not use this factory directly, instead get an instance of
    the factory from the `pygw.query.query_builder.QueryBuilder.constraints_factory` method of the query builder.
    """

    def data_ids(self, data_ids):
        """
        Constrain a query by data IDs.

        Args:
            data_ids (list of bytes): The data IDs to constrain by.
        Returns:
            A `pygw.query.query_constraints.QueryConstraints` with the given data ids.
        """
        byte_array_class = JavaClass("[B", java_gateway._gateway_client)
        j_data_ids = java_gateway.new_array(byte_array_class, len(data_ids))
        for idx, data_id in enumerate(data_ids):
            j_data_ids[idx] = _pbat.to_java(data_id)
        j_qc = self._java_ref.dataIds(j_data_ids)
        return QueryConstraints(j_qc)

    def data_ids_by_range(self, start_data_id_inclusive, end_data_id_inclusive):
        """
        Constrain a query using a range of data IDs, assuming big endian ordering.

        Args:
            start_data_id_inclusive (bytes): The start of data ID range (inclusive).
            end_data_id_inclusive (bytes): The end of data ID range (inclusive).
        Returns:
            A `pygw.query.query_constraints.QueryConstraints` with the given data ID range.
        """
        j_qc = self._java_ref.dataIdsByRange(_pbat.to_java(start_data_id_inclusive),
                                             _pbat.to_java(end_data_id_inclusive))
        return QueryConstraints(j_qc)

    def prefix(self, partition_key, sort_key_prefix):
        """
        Constrain a query by prefix.

        Args:
            partition_key (bytes): The partition key to constrain by.
            sort_key_prefix (bytes): The sort key prefix to constrain by.
        Returns:
            A `pygw.query.query_constraints.QueryConstraints` with the given prefix.
        """
        j_qc = self._java_ref.prefix(_pbat.to_java(partition_key),
                                     _pbat.to_java(sort_key_prefix))
        return QueryConstraints(j_qc)

    def coordinate_ranges(self, index_strategy, coordinate_ranges):
        # TODO: support Python variants of NumericIndexStrategy and MultiDimensionalCoordinateRangesArray
        raise NotImplementedError

    def constraints(self, constraints, compare_op=None):
        # TODO: support Python variants of Constraints
        raise NotImplementedError

    def no_constraints(self):
        """
        Use no query constraints, meaning wide open query (this is the default).

        Returns:
            A `pygw.query.query_constraints.QueryConstraints` with no constraints.
        """
        return QueryConstraints(self._java_ref.noConstraints())



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.config import geowave_pkg

from .index_builder import IndexBuilder


class SpatialIndexBuilder(IndexBuilder):
    """
    Index builder for a spatial index.
    """

    def __init__(self):
        super().__init__(geowave_pkg.core.geotime.index.api.SpatialIndexBuilder())

    def set_include_time_in_common_index_model(self, include):
        """
        Sets whether or not to include time in the common index model.  This can be
        used to speed up queries that may involve temporal constraints.

        Args:
            include (bool): Whether or not to include time in the common index model.
        Returns:
            This index builder.
        """
        self._java_ref.setIncludeTimeInCommonIndexModel(include)
        return self



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.base import GeoWaveObject
from .statistic_mappings import map_statistic
from ..base.java_transformer import JavaTransformer


class StatisticValue(GeoWaveObject):
    """
    Base GeoWave statistic value.
    """

    def __init__(self, java_ref):
        super().__init__(java_ref)

    def get_statistic(self):
        """
        Get the parent statistic.

        Returns:
            The statistic associated with this value
        """
        return map_statistic(self._java_ref.getStatistic())

    def get_bin(self):
        """
        Gets the bin for this value. If the underlying statistic does not use a binning strategy, an empty byte array
        will be returned.

        Returns:
            The bin for this value.
        """
        return self._java_ref.getBin()

    def get_value(self):
        """
        Gets the raw statistic value.

        Returns:
            The raw statistic value.
        """
        statistic = self.get_statistic()
        return statistic.java_transformer.transform(self._java_ref.getValue())

    def merge(self, other):
        """
        Merge another statistic value of the same type into this statistic value.

        Args:
            other (StatisticValue): The other value to merge into this one.
        """
        self._java_ref.merge(other.java_ref())


class StatisticValueTransformer(JavaTransformer):
    def transform(self, j_object):
        return StatisticValue(j_object)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
"""
This module contains field statistics

It contains the following import shortcuts:
```python
from pygw.statistics.field import BloomFilterStatistic
from pygw.statistics.field import BoundingBoxStatistic
from pygw.statistics.field import CountMinSketchStatistic
from pygw.statistics.field import FixedBinNumericHistogramStatistic
from pygw.statistics.field import HyperLogLogStatistic
from pygw.statistics.field import NumericHistogramStatistic
from pygw.statistics.field import NumericMeanStatistic
from pygw.statistics.field import NumericRangeStatistic
from pygw.statistics.field import NumericStatsStatistic
from pygw.statistics.field import TimeRangeStatistic
```
"""

from .bloom_filter_statistic import BloomFilterStatistic
from .bounding_box_statistic import BoundingBoxStatistic
from .count_min_sketch_statistic import CountMinSketchStatistic
from .fixed_bin_numeric_histogram_statistic import FixedBinNumericHistogramStatistic
from .hyper_log_log_statistic import HyperLogLogStatistic
from .numeric_histogram_statistic import NumericHistogramStatistic
from .numeric_mean_statistic import NumericMeanStatistic
from .numeric_range_statistic import NumericRangeStatistic
from .numeric_stats_statistic import NumericStatsStatistic
from .time_range_statistic import TimeRangeStatistic



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

import pytest
import os

from pygw.store import DataStoreFactory
from pygw.store.accumulo import AccumuloOptions
from pygw.store.bigtable import BigTableOptions
from pygw.store.cassandra import CassandraOptions
from pygw.store.dynamodb import DynamoDBOptions
from pygw.store.hbase import HBaseOptions
from pygw.store.kudu import KuduOptions
from pygw.store.redis import RedisOptions
from pygw.store.rocksdb import RocksDBOptions
from pygw.index import SpatialIndexBuilder
from pygw.query import VectorQueryBuilder

from .conftest import TEST_DIR, POINT_NUMBER_FIELD, POINT_TYPE_NAME
from .conftest import POINT_TYPE_ADAPTER
from .conftest import TEST_DATA
from .conftest import write_test_data
from .conftest import results_as_list


# Test Additions #
from ..statistics.data_type import CountStatistic
from ..statistics.field import NumericRangeStatistic
from ..statistics.index import DuplicateEntryCountStatistic


def test_add_type(test_ds):
    # given
    index = SpatialIndexBuilder().create_index()
    adapter = POINT_TYPE_ADAPTER

    # when
    test_ds.add_type(adapter, index)
    indices = test_ds.get_indices()
    types = test_ds.get_types()

    # then
    assert len(indices) == 1
    assert indices[0].get_name() == index.get_name()
    assert indices[0].get_index_strategy() == index.get_index_strategy()
    assert indices[0].get_index_model() == index.get_index_model()
    assert len(types) == 1
    assert types[0].get_type_name() == adapter.get_type_name()


def test_add_existing_type(test_ds):
    # given
    index = SpatialIndexBuilder().create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    test_ds.add_type(adapter, index)
    indices = test_ds.get_indices(adapter.get_type_name())

    # then
    assert len(indices) == 1
    assert indices[0].get_name() == index.get_name()
    assert indices[0].get_index_strategy() == index.get_index_strategy()
    assert indices[0].get_index_model() == index.get_index_model()


# Test Removing #
def test_remove_index(test_ds):
    # given
    index = SpatialIndexBuilder().set_name("idx1").create_index()
    index2 = SpatialIndexBuilder().set_name("idx2").create_index()

    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)
    test_ds.add_type(adapter, index2)

    # when
    test_ds.remove_index(adapter.get_type_name(), index.get_name())
    indices = test_ds.get_indices()

    # then
    assert len(indices) == 1
    assert indices[0].get_name() == index2.get_name()
    assert indices[0].get_index_strategy() == index2.get_index_strategy()
    assert indices[0].get_index_model() == index2.get_index_model()


def test_remove_index_last(test_ds):
    with pytest.raises(Exception) as exec:
        # given
        index = SpatialIndexBuilder().create_index()
        adapter = POINT_TYPE_ADAPTER
        test_ds.add_type(adapter, index)

        # when
        test_ds.remove_index(index.get_name())

    # then
    assert 'Adapters require at least one index' in str(exec.value)


def test_remove_index_non_exist(test_ds):
    # given
    index = SpatialIndexBuilder().create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    test_ds.remove_index("Corgi")

    # then
    assert len(test_ds.get_indices()) == 1


def test_remove_type(test_ds):
    # given
    index = SpatialIndexBuilder().create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)
    write_test_data(test_ds, index)

    # when
    test_ds.remove_type(adapter.get_type_name())
    query = VectorQueryBuilder().build()
    res = results_as_list(test_ds.query(query))

    # then
    assert len(test_ds.get_indices(adapter.get_type_name())) == 0
    assert len(test_ds.get_indices()) == 1
    assert len(res) == 0


# Test Deleting #
def test_delete(test_ds):
    # given
    index = SpatialIndexBuilder().set_name("idx1").create_index()
    index2 = SpatialIndexBuilder().set_name("idx2").create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)
    test_ds.add_type(adapter, index2)
    write_test_data(test_ds, index, index2)

    # when
    qbldr = VectorQueryBuilder()
    constraints_factory = qbldr.constraints_factory()
    # filter encompasses 10 features (1, 1) - (10, 10)
    constraints = constraints_factory.cql_constraints("BBOX(the_geom, 0.5, 0.5, 10.5, 10.5)")
    qbldr.constraints(constraints)
    test_ds.delete(qbldr.build())

    query = VectorQueryBuilder().build()
    res = results_as_list(test_ds.query(query))

    # then
    assert len(test_ds.get_indices()) == 2
    assert len(test_ds.get_types()) == 1
    assert len(res) == (len(TEST_DATA) - 10)


# Test Delete All #
def test_delete_all(test_ds):
    # given
    index = SpatialIndexBuilder().set_name("idx1").create_index()
    index2 = SpatialIndexBuilder().set_name("idx2").create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)
    test_ds.add_type(adapter, index2)
    write_test_data(test_ds, index, index2)

    # when
    test_ds.delete_all()

    query = VectorQueryBuilder().build()
    res = results_as_list(test_ds.query(query))

    # then
    assert len(test_ds.get_indices()) == 0
    assert len(test_ds.get_types()) == 0
    assert len(res) == 0


# Test Copy #
def test_copy(test_ds):
    # given
    options = RocksDBOptions()
    options.set_geowave_namespace("geowave.tests")
    options.set_directory(os.path.join(TEST_DIR, "datastore2"))
    ds2 = DataStoreFactory.create_data_store(options)

    adapter = POINT_TYPE_ADAPTER
    index = SpatialIndexBuilder().create_index()
    test_ds.add_type(adapter, index)

    write_test_data(test_ds, index)

    # when
    test_ds.copy_to(ds2)

    indices = ds2.get_indices()
    types = ds2.get_types()
    query = VectorQueryBuilder().build()
    res = results_as_list(ds2.query(query))

    # then
    assert len(test_ds.get_indices()) == 1
    assert len(indices) == 1
    assert indices[0].get_name() == index.get_name()
    assert indices[0].get_index_strategy() == index.get_index_strategy()
    assert indices[0].get_index_model() == index.get_index_model()
    assert len(types) == 1
    assert types[0].get_type_name() == adapter.get_type_name()
    assert len(res) == len(TEST_DATA)

    ds2.delete_all()


def test_copy_by_query(test_ds):
    # given
    options = RocksDBOptions()
    options.set_geowave_namespace("geowave.tests")
    options.set_directory(os.path.join(TEST_DIR, "datastore2"))
    ds2 = DataStoreFactory.create_data_store(options)

    adapter = POINT_TYPE_ADAPTER
    index = SpatialIndexBuilder().create_index()
    test_ds.add_type(adapter, index)

    write_test_data(test_ds, index)

    # when
    qbldr = VectorQueryBuilder()
    constraints_factory = qbldr.constraints_factory()
    # filter encompasses 10 features (1, 1) - (10, 10)
    constraints = constraints_factory.cql_constraints("BBOX(the_geom, 0.5, 0.5, 10.5, 10.5)")
    qbldr.all_indices().constraints(constraints)
    test_ds.copy_to(ds2, qbldr.build())

    indices = ds2.get_indices()
    types = ds2.get_types()
    query = VectorQueryBuilder().build()
    res = results_as_list(ds2.query(query))

    # then
    assert len(test_ds.get_indices()) == 1
    assert len(indices) == 1
    assert indices[0].get_name() == index.get_name()
    assert indices[0].get_index_strategy() == index.get_index_strategy()
    assert indices[0].get_index_model() == index.get_index_model()
    assert len(types) == 1
    assert types[0].get_type_name() == adapter.get_type_name()
    assert len(res) == 10

    ds2.delete_all()


# Test Writer #
def test_create_writer(test_ds):
    # given
    adapter = POINT_TYPE_ADAPTER
    index = SpatialIndexBuilder().create_index()
    test_ds.add_type(adapter, index)

    # when
    writer = test_ds.create_writer(adapter.get_type_name())

    # then
    assert writer is not None


def test_create_writer_null(test_ds):
    # when
    writer = test_ds.create_writer("Corgi")

    # then
    assert writer is None


def test_create_writer_null_other(test_ds):
    # given
    adapter = POINT_TYPE_ADAPTER
    index = SpatialIndexBuilder().create_index()
    test_ds.add_type(adapter, index)
    test_ds.create_writer(adapter.get_type_name())

    # when
    writer = test_ds.create_writer("Corgi")

    # then
    assert writer is None


def test_write(test_ds):
    # given
    adapter = POINT_TYPE_ADAPTER
    index = SpatialIndexBuilder().create_index()
    test_ds.add_type(adapter, index)

    # when
    write_test_data(test_ds, index)

    query = VectorQueryBuilder().build()

    res = results_as_list(test_ds.query(query))

    # then
    assert len(res) == len(TEST_DATA)


def test_add_get_statistics(test_ds):
    # given
    adapter = POINT_TYPE_ADAPTER
    index = SpatialIndexBuilder().set_name('idx').create_index()
    test_ds.add_type(adapter, index)
    write_test_data(test_ds, index)

    duplicate_entry_count_stat = DuplicateEntryCountStatistic('idx')
    duplicate_entry_count_stat.set_tag('test')
    count_stat = CountStatistic(POINT_TYPE_NAME)
    count_stat.set_tag('test')
    numeric_range_statistic = NumericRangeStatistic(POINT_TYPE_NAME, POINT_NUMBER_FIELD)
    numeric_range_statistic.set_tag('test')

    # when
    test_ds.add_statistic(duplicate_entry_count_stat, count_stat, numeric_range_statistic)

    # then
    duplicate_entry_count_stat = test_ds.get_index_statistic(DuplicateEntryCountStatistic.STATS_TYPE, 'idx', 'test')
    assert isinstance(duplicate_entry_count_stat, DuplicateEntryCountStatistic)
    assert duplicate_entry_count_stat.get_tag() == 'test'
    assert duplicate_entry_count_stat.get_index_name() == 'idx'
    count_stat = test_ds.get_data_type_statistic(CountStatistic.STATS_TYPE, POINT_TYPE_NAME, 'test')
    assert isinstance(count_stat, CountStatistic)
    assert count_stat.get_tag() == 'test'
    assert count_stat.get_type_name() == POINT_TYPE_NAME
    numeric_range_statistic = test_ds.get_field_statistic(
        NumericRangeStatistic.STATS_TYPE, POINT_TYPE_NAME, POINT_NUMBER_FIELD, 'test')
    assert isinstance(numeric_range_statistic, NumericRangeStatistic)
    assert numeric_range_statistic.get_tag() == 'test'
    assert numeric_range_statistic.get_type_name() == POINT_TYPE_NAME
    assert numeric_range_statistic.get_field_name() == POINT_NUMBER_FIELD

    index_stats = test_ds.get_index_statistics('idx')
    assert any(stat.get_tag() == 'test' and isinstance(stat, DuplicateEntryCountStatistic) for stat in index_stats)
    data_type_stats = test_ds.get_data_type_statistics(POINT_TYPE_NAME)
    assert any(stat.get_tag() == 'test' and isinstance(stat, CountStatistic) for stat in data_type_stats)
    field_stats = test_ds.get_field_statistics(POINT_TYPE_NAME, POINT_NUMBER_FIELD)
    assert any(stat.get_tag() == 'test' and isinstance(stat, NumericRangeStatistic) for stat in field_stats)


def test_stat_recalc_statistics(test_ds):
    # given
    adapter = POINT_TYPE_ADAPTER
    index = SpatialIndexBuilder().set_name('idx').create_index()
    test_ds.add_type(adapter, index)
    write_test_data(test_ds, index)

    duplicate_entry_count_stat = DuplicateEntryCountStatistic('idx')
    duplicate_entry_count_stat.set_tag('test')
    count_stat = CountStatistic(POINT_TYPE_NAME)
    count_stat.set_tag('test')
    numeric_range_statistic = NumericRangeStatistic(POINT_TYPE_NAME, POINT_NUMBER_FIELD)
    numeric_range_statistic.set_tag('test')
    test_ds.add_empty_statistic(duplicate_entry_count_stat, count_stat, numeric_range_statistic)
    assert test_ds.get_statistic_value(duplicate_entry_count_stat) == 0
    assert test_ds.get_statistic_value(count_stat) == 0
    numeric_range = test_ds.get_statistic_value(numeric_range_statistic)
    assert numeric_range.get_minimum() == 0
    assert numeric_range.get_maximum() == 0

    # when
    test_ds.recalc_statistic(duplicate_entry_count_stat, count_stat, numeric_range_statistic)

    # then
    assert test_ds.get_statistic_value(duplicate_entry_count_stat) == 0
    assert test_ds.get_statistic_value(count_stat) == 360
    numeric_range = test_ds.get_statistic_value(numeric_range_statistic)
    assert numeric_range.get_minimum() == -180
    assert numeric_range.get_maximum() == 179

def test_remove_statistics(test_ds):
    # given
    adapter = POINT_TYPE_ADAPTER
    index = SpatialIndexBuilder().set_name('idx').create_index()
    test_ds.add_type(adapter, index)
    write_test_data(test_ds, index)

    duplicate_entry_count_stat = DuplicateEntryCountStatistic('idx')
    duplicate_entry_count_stat.set_tag('test')
    count_stat = CountStatistic(POINT_TYPE_NAME)
    count_stat.set_tag('test')
    numeric_range_statistic = NumericRangeStatistic(POINT_TYPE_NAME, POINT_NUMBER_FIELD)
    numeric_range_statistic.set_tag('test')

    test_ds.add_statistic(duplicate_entry_count_stat, count_stat, numeric_range_statistic)

    # and
    duplicate_entry_count_stat = test_ds.get_index_statistic(DuplicateEntryCountStatistic.STATS_TYPE, 'idx', 'test')
    assert isinstance(duplicate_entry_count_stat, DuplicateEntryCountStatistic)
    assert duplicate_entry_count_stat.get_tag() == 'test'
    assert duplicate_entry_count_stat.get_index_name() == 'idx'
    count_stat = test_ds.get_data_type_statistic(CountStatistic.STATS_TYPE, POINT_TYPE_NAME, 'test')
    assert isinstance(count_stat, CountStatistic)
    assert count_stat.get_tag() == 'test'
    assert count_stat.get_type_name() == POINT_TYPE_NAME
    numeric_range_statistic = test_ds.get_field_statistic(
        NumericRangeStatistic.STATS_TYPE, POINT_TYPE_NAME, POINT_NUMBER_FIELD, 'test')
    assert isinstance(numeric_range_statistic, NumericRangeStatistic)
    assert numeric_range_statistic.get_tag() == 'test'
    assert numeric_range_statistic.get_type_name() == POINT_TYPE_NAME
    assert numeric_range_statistic.get_field_name() == POINT_NUMBER_FIELD

    # when
    test_ds.remove_statistic(duplicate_entry_count_stat, count_stat, numeric_range_statistic)

    # then
    assert test_ds.get_index_statistic(DuplicateEntryCountStatistic.STATS_TYPE, 'idx', 'test') is None
    assert test_ds.get_data_type_statistic(CountStatistic.STATS_TYPE, POINT_TYPE_NAME, 'test') is None
    assert test_ds.get_field_statistic(
        NumericRangeStatistic.STATS_TYPE, POINT_TYPE_NAME, POINT_NUMBER_FIELD, 'test') is None


def _test_base_options(options, server_side_possible=True):
    options.set_geowave_namespace("test_namespace")
    assert options.get_geowave_namespace() == "test_namespace"

    persist_data_statistics = not options.is_persist_data_statistics()
    options.set_persist_data_statistics(persist_data_statistics)
    assert options.is_persist_data_statistics() == persist_data_statistics

    secondary_indexing = not options.is_secondary_indexing()
    options.set_secondary_indexing(secondary_indexing)
    assert options.is_secondary_indexing() == secondary_indexing

    block_cache = not options.is_enable_block_cache()
    options.set_enable_block_cache(block_cache)
    assert options.is_enable_block_cache() == block_cache

    server_side_library = not options.is_server_side_library_enabled()
    options.set_secondary_indexing(False)
    options.set_server_side_library_enabled(server_side_library)
    if server_side_possible:
        assert options.is_server_side_library_enabled() == server_side_library
    else:
        assert options.is_server_side_library_enabled() is False

    options.set_max_range_decomposition(42)
    assert options.get_max_range_decomposition() == 42

    options.set_aggregation_max_range_decomposition(43)
    assert options.get_aggregation_max_range_decomposition() == 43

    visibility = not options.is_visibility_enabled()
    options.set_enable_visibility(visibility)
    assert options.is_visibility_enabled() == visibility


def test_accumulo_options():
    options = AccumuloOptions()
    options.set_zookeeper("test_zookeeper")
    assert options.get_zookeeper() == "test_zookeeper"
    options.set_instance("test_instance")
    assert options.get_instance() == "test_instance"
    options.set_user("test_user")
    assert options.get_user() == "test_user"
    options.set_password("test_password")
    assert options.get_password() == "test_password"
    locality_groups = not options.is_use_locality_groups()
    options.set_use_locality_groups(locality_groups)
    assert options.is_use_locality_groups() == locality_groups
    _test_base_options(options)


def test_bigtable_options():
    options = BigTableOptions()
    options.set_scan_cache_size(42)
    assert options.get_scan_cache_size() == 42
    options.set_project_id("test_project_id")
    assert options.get_project_id() == "test_project_id"
    options.set_instance_id("test_instance_id")
    assert options.get_instance_id() == "test_instance_id"
    _test_base_options(options)


def test_cassandra_options():
    options = CassandraOptions()
    options.set_contact_points("test_contact_point")
    assert options.get_contact_points() == "test_contact_point"
    options.set_datacenter("test_datacenter")
    assert options.get_datacenter() == "test_datacenter"
    options.set_batch_write_size(42)
    assert options.get_batch_write_size() == 42
    durable_writes = not options.is_durable_writes()
    options.set_durable_writes(durable_writes)
    assert options.is_durable_writes() == durable_writes
    options.set_replication_factor(43)
    assert options.get_replication_factor() == 43    
    options.set_gc_grace_seconds(44)
    assert options.get_gc_grace_seconds() == 44
    table_options = {
        "test_key_1": "test_value_1",
        "test_key_2": "test_value_2"
    }
    options.set_table_options(table_options)
    returned_table_options = options.get_table_options()
    assert len(returned_table_options) == len(table_options)
    for key in returned_table_options:
        assert (key in table_options and returned_table_options[key] == table_options[key])
    options.set_compaction_strategy("TimeWindowCompactionStrategy")
    assert options.get_compaction_strategy() == "TimeWindowCompactionStrategy"
    _test_base_options(options, False)


def test_dynamodb_options():
    options = DynamoDBOptions()
    options.set_region("us-east-1")
    assert options.get_region() == "us-east-1"
    options.set_region(None)
    assert options.get_region() is None
    options.set_endpoint("test_endpoint")
    assert options.get_endpoint() == "test_endpoint"
    options.set_write_capacity(42)
    assert options.get_write_capacity() == 42
    options.set_read_capacity(43)
    assert options.get_read_capacity() == 43
    enable_cache_response_metadata = not options.is_enable_cache_response_metadata()
    options.set_enable_cache_response_metadata(enable_cache_response_metadata)
    assert options.is_enable_cache_response_metadata() == enable_cache_response_metadata
    options.set_protocol("HTTP")
    assert options.get_protocol() == "HTTP"
    options.set_protocol(None)
    assert options.get_protocol() is None
    options.set_max_connections(44)
    assert options.get_max_connections() == 44
    _test_base_options(options, False)


def test_hbase_options():
    options = HBaseOptions()
    options.set_zookeeper("test_zookeeper")
    assert options.get_zookeeper() == "test_zookeeper"
    options.set_scan_cache_size(42)
    assert options.get_scan_cache_size() == 42
    options.set_server_side_library_enabled(True)
    verify_coprocessors = not options.is_verify_coprocessors()
    options.set_verify_coprocessors(verify_coprocessors)
    assert options.is_verify_coprocessors() == verify_coprocessors
    options.set_coprocessor_jar("test_jar")
    assert options.get_coprocessor_jar() == "test_jar"
    _test_base_options(options)


def test_kudu_options():
    options = KuduOptions()
    options.set_kudu_master("test_master")
    assert options.get_kudu_master() == "test_master"
    _test_base_options(options, False)


def test_redis_options():
    options = RedisOptions()
    options.set_address("test_address")
    assert options.get_address() == "test_address"
    options.set_compression("L4Z")
    assert options.get_compression() == "L4Z"
    options.set_username("test_username")
    assert options.get_username() == "test_username"
    options.set_password("test_password")
    assert options.get_password() == "test_password"
    _test_base_options(options, False)


def test_rocksdb_options():
    options = RocksDBOptions()
    options.set_directory("test_directory")
    assert options.get_directory() == "test_directory"
    compact_on_wriite = not options.is_compact_on_write()
    options.set_compact_on_write(compact_on_wriite)
    assert options.is_compact_on_write() == compact_on_wriite
    options.set_batch_write_size(42)
    assert options.get_batch_write_size() == 42
    _test_base_options(options, False)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.config import geowave_pkg
from ...base.envelope import EnvelopeTransformer
from ...base.interval import IntervalTransformer

from ...base.type_conversions import StringArrayType
from ..aggregation_query_builder import AggregationQueryBuilder
from .vector_query_constraints_factory import VectorQueryConstraintsFactory


class VectorAggregationQueryBuilder(AggregationQueryBuilder):
    """
    A builder for creating aggregation queries for vector data.
    """

    def __init__(self):
        j_agg_qbuilder = geowave_pkg.core.geotime.store.query.api.VectorAggregationQueryBuilder.newBuilder()
        super().__init__(j_agg_qbuilder)

    def constraints_factory(self):
        """
        Creates a constraints factory for vector queries.  The vector query constraint factory
        provides additional constraints specific to vector data.

        Returns:
            A `pygw.query.vector.VectorQueryConstraintsFactory`.
        """
        return VectorQueryConstraintsFactory(self._java_ref.constraintsFactory())

    def bbox_of_results(self, *type_names):
        """
        Convenience method for getting a bounding box of the results of a query. It uses the default geometry for a
        feature type which is also the indexed geometry.

        Args:
            type_names (str): The type names to get the bounding box of.
        Returns:
            This query builder.
        """
        self._java_transformer = EnvelopeTransformer()
        if type_names is None:
            self._java_ref.bboxOfResults()
        else:
            self._java_ref.bboxOfResults(StringArrayType().to_java(type_names))
        return self

    def bbox_of_results_for_geometry_field(self, type_name, geometry_field_name):
        """
        Convenience method for getting a bounding box of the results of a query.  This can be particularly useful if you
        want to calculate the bbox on a different field than the default/indexed Geometry.

        Args:
            type_name (str): The type to aggregate.
            geometry_field_name (str): The geometry field to get the bounding box of.
        Returns:
            This query builder.
        """
        self._java_transformer = EnvelopeTransformer()
        self._java_ref.bboxOfResultsForGeometryField(type_name, geometry_field_name)
        return self

    def time_range_of_results(self, *type_names):
        """
        Convenience method for getting a time range of the results of a query.  This uses inferred or hinted
        temporal attribute names.

        Args:
            type_names (str): The type names to get the time range of.
        Returns:
            This query builder.
        """
        self._java_transformer = IntervalTransformer()
        self._java_ref.timeRangeOfResults(StringArrayType().to_java(type_names))
        return self

    def time_range_of_results_for_time_field(self, type_name, time_field_name):
        """
        Convenience method for getting a time range of the results of a query.  This can be particularly useful if you
        want to calculate the time range on a specific time field.

        Args:
            type_name (str): The type to aggregate.
            time_field_name (str): The time field to get the range of.
        Returns:
            This query builder.
        """
        self._java_transformer = IntervalTransformer()
        self._java_ref.timeRangeOfResultsForTimeField(type_name, time_field_name)
        return self

    def min(self, type_name, numeric_field_name):
        """
        Convenience method for getting the minimum value of a numeric field from the results of a query.

        Args:
            type_name (str): The type to aggregate.
            numeric_field_name (str): The numeric field to get the minimum value of.
        Returns:
            This query builder.
        """
        j_field_parameter = geowave_pkg.core.store.query.aggregate.FieldNameParam(numeric_field_name)
        j_min_agg = geowave_pkg.core.store.query.aggregate.FieldMinAggregation(j_field_parameter)
        return self.aggregate(type_name, j_min_agg)

    def max(self, type_name, numeric_field_name):
        """
        Convenience method for getting the maximum value of a numeric field from the results of a query.

        Args:
            type_name (str): The type to aggregate.
            numeric_field_name (str): The numeric field to get the maximum value of.
        Returns:
            This query builder.
        """
        j_field_parameter = geowave_pkg.core.store.query.aggregate.FieldNameParam(numeric_field_name)
        j_max_agg = geowave_pkg.core.store.query.aggregate.FieldMaxAggregation(j_field_parameter)
        return self.aggregate(type_name, j_max_agg)

    def sum(self, type_name, numeric_field_name):
        """
        Convenience method for getting the sum of a numeric field from the results of a query.

        Args:
            type_name (str): The type to aggregate.
            numeric_field_name (str): The numeric field to get the sum of.
        Returns:
            This query builder.
        """
        j_field_parameter = geowave_pkg.core.store.query.aggregate.FieldNameParam(numeric_field_name)
        j_sum_agg = geowave_pkg.core.store.query.aggregate.FieldSumAggregation(j_field_parameter)
        return self.aggregate(type_name, j_sum_agg)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
"""
The `config` module includes several shortcut objects to make accessing the Py4J Java gateway more convenient.  These
include:
- *`java_gateway`* Py4J Gateway Object
- *`java_pkg`*: Shortcut for `java_gateway.jvm`.  Can be used to construct JVM objects like
      `java_pkg.org.geotools.feature.simple.SimpleFeatureTypeBuilder()`
- *`geowave_pkg`*: Similar to `java_pkg`, serves as a shortcut for `java_gateway.jvm.org.locationtech.geowave`.
- *`reflection_util`*: Direct access to the Py4J reflection utility.

These objects can be imported directly using `from pygw.config import <object_name>`.

This module uses the `gateway` module to connect to the Py4J Java gateway.  By default, the gateway module will attempt
to connect to a locally running gateway.  See the documentation of that module for information about configuring `pygw`
to use other gateways.
"""
from pygw.gateway import gateway_config

gateway_config.init()

java_gateway = gateway_config.GATEWAY
"""py4j.java_gateway.JavaGateway: The gateway between pygw and the JVM."""

java_pkg = gateway_config.GATEWAY.jvm
"""py4j.java_gateway.JVMView: A shortcut for accessing java packages directly.

For example `java_pkg.org.geotools.feature.simple.SimpleFeatureTypeBuilder`.
"""

geowave_pkg = gateway_config.GATEWAY.jvm.org.locationtech.geowave
"""py4j.java_gateway.JVMView: A shortcut for accessing geowave packages directly.

For example `geowave_pkg.core.store.api.DataStoreFactory`.
"""

reflection_util = gateway_config.GATEWAY.jvm.py4j.reflection.ReflectionUtil
"""py4j.java_gateway.JavaClass: A Java reflection utility."""

__all__ = ["java_gateway", "java_pkg", "geowave_pkg", "reflection_util"]



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.config import geowave_pkg
from ..statistic import IndexStatistic
from ..statistic_type import IndexStatisticType
from ...base.java_transformer import JavaTransformer
from ...base.type_conversions import PrimitiveByteArrayType


class PartitionsStatistic(IndexStatistic):
    """
    This class is responsible for maintaining all unique Partition IDs that are being used within a data set.
    """
    STATS_TYPE = IndexStatisticType(geowave_pkg.core.store.statistics.index.PartitionsStatistic.STATS_TYPE)

    def __init__(self, index_name=None, java_ref=None):
        if java_ref is None:
            if index_name is None:
                java_ref = geowave_pkg.core.store.statistics.index.PartitionsStatistic()
            else:
                java_ref = geowave_pkg.core.store.statistics.index.PartitionsStatistic(index_name)
        super().__init__(java_ref, PartitionsTransformer())


class PartitionsTransformer(JavaTransformer):
    def transform(self, j_object):
        partitions = set()
        byte_array_type = PrimitiveByteArrayType()
        for item in j_object:
            partitions.add(byte_array_type.from_java(item.getBytes()))
        return partitions



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
"""
This module contains classes that can be used to establish connections to the various GeoWave backends.  Each store type
has a submodule which contains a class that can be used to connect to that store type.  For example
`from pygw.store.accumulo import AccumuloOptions`.  The `DataStore` object can be constructed by passing the options
object to the `DataStoreFactory.create_data_store(<options>)` method.

This module contains the following import shortcuts:
```python
from pygw.store import DataStore
from pygw.store import DataStoreOptions
from pygw.store import DataStoreFactory
```
"""

from .data_store import DataStore
from .data_store_options import DataStoreOptions
from .data_store_factory import DataStoreFactory



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.config import geowave_pkg
from ..statistic import FieldStatistic
from ..statistic_type import FieldStatisticType
from ...base import GeoWaveObject
from ...base.java_transformer import JavaTransformer


class NumericHistogramStatistic(FieldStatistic):
    """
    Dynamic histogram provide very high accuracy for CDF and quantiles over the a numeric attribute.
    """
    STATS_TYPE = FieldStatisticType(geowave_pkg.core.store.statistics.field.NumericHistogramStatistic.STATS_TYPE)

    def __init__(self, type_name=None, field_name=None, compression=100, java_ref=None):
        if java_ref is None:
            if type_name is None and field_name is None:
                java_ref = geowave_pkg.core.store.statistics.field.NumericHistogramStatistic()
            else:
                java_ref = geowave_pkg.core.store.statistics.field.NumericHistogramStatistic(
                    type_name, field_name, float(compression))
        super().__init__(java_ref, NumericHistogramTransformer())

    def set_compression(self, compression):
        self._java_ref.setCompression(float(compression))

    def get_compression(self):
        return self._java_ref.getCompression()


class NumericHistogramTransformer(JavaTransformer):
    def transform(self, j_object):
        return NumericHistogram(j_object)


class NumericHistogram(GeoWaveObject):
    def quantile(self, value):
        return self._java_ref.quantile(float(value))

    def cdf(self, value):
        return self._java_ref.cdf(float(value))

    def sum(self, value, inclusive=True):
        return self._java_ref.sum(float(value), inclusive)

    def get_min_value(self):
        return self._java_ref.getMinValue()

    def get_max_value(self):
        return self._java_ref.getMaxValue()

    def get_total_count(self):
        return self._java_ref.getTotalCount()



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.base import GeoWaveObject

from .index import Index


class IndexBuilder(GeoWaveObject):
    """
    Base class for building indices.
    """

    def set_num_partitions(self, num_partitions):
        """
        Sets the number of partitions for the index to use.

        Args:
            num_partitions (int): The number of partitions to use.
        Returns:
            This index builder.
        """
        self._java_ref.setNumPartitions(num_partitions)
        return self

    def set_name(self, index_name):
        """
        Set the name of the index to the given value.

        Args:
            index_name (str): The name to use.
        Returns:
            This index builder.
        """
        self._java_ref.setName(index_name)
        return self

    def create_index(self):
        """
        Builds the configured index.

        Returns:
            A `pygw.index.index.Index` with the given configuration.
        """
        j_idx = self._java_ref.createIndex()
        return Index(j_idx)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.config import geowave_pkg
from ..statistic import IndexStatistic
from ..statistic_type import IndexStatisticType


class MaxDuplicatesStatistic(IndexStatistic):
    """
    Maintains the maximum number of duplicates that a single entry in the data set contains.
    """
    STATS_TYPE = IndexStatisticType(geowave_pkg.core.store.statistics.index.MaxDuplicatesStatistic.STATS_TYPE)

    def __init__(self, index_name=None, java_ref=None):
        if java_ref is None:
            if index_name is None:
                java_ref = geowave_pkg.core.store.statistics.index.MaxDuplicatesStatistic()
            else:
                java_ref = geowave_pkg.core.store.statistics.index.MaxDuplicatesStatistic(index_name)
        super().__init__(java_ref)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.config import geowave_pkg
from ..statistic import FieldStatistic
from ..statistic_type import FieldStatisticType
from ...base import GeoWaveObject
from ...base.java_transformer import JavaTransformer
from ...base.type_conversions import PrimitiveDoubleArrayType, PrimitiveLongArrayType


class FixedBinNumericHistogramStatistic(FieldStatistic):
    """
    Fixed number of bins for a histogram. Unless configured, the range will expand dynamically,  redistributing the
    data as necessary into the wider bins.

    The advantage of constraining the range of the statistic is to ignore values outside the range, such as erroneous
    values. Erroneous values force extremes in the histogram. For example, if the expected range of values falls
    between 0 and 1 and a value of 10000 occurs, then a single bin contains the entire population between 0 and 1, a
    single bin represents the single value of 10000.  If there are extremes in the data, then use
    NumericHistogramStatistic instead.
    """
    STATS_TYPE = FieldStatisticType(
        geowave_pkg.core.store.statistics.field.FixedBinNumericHistogramStatistic.STATS_TYPE)

    def __init__(self, type_name=None, field_name=None, bins=1024, min_value=None, max_value=None, java_ref=None):
        if java_ref is None:
            if type_name is None and field_name is None:
                java_ref = geowave_pkg.core.store.statistics.field.FixedBinNumericHistogramStatistic()
            else:
                java_ref = geowave_pkg.core.store.statistics.field.FixedBinNumericHistogramStatistic(
                    type_name, field_name, bins, float(min_value), float(max_value))
        super().__init__(java_ref, FixedBinNumericHistogramTransformer())

    def set_num_bins(self, num_bins):
        self._java_ref.setNumBins(num_bins)

    def get_num_bins(self):
        return self._java_ref.getNumBins()

    def set_min_value(self, min_value):
        self._java_ref.setMinValue(float(min_value))

    def get_min_value(self):
        return self._java_ref.getMinValue()

    def set_max_value(self, max_value):
        self._java_ref.setMaxValue(float(max_value))

    def get_max_value(self):
        return self._java_ref.getMaxValue()


class FixedBinNumericHistogramTransformer(JavaTransformer):
    def transform(self, j_object):
        return FixedBinNumericHistogram(j_object)


class FixedBinNumericHistogram(GeoWaveObject):
    def bin_quantiles(self, bins):
        return PrimitiveDoubleArrayType().from_java(self._java_ref.quantile(int(bins)))

    def quantile(self, percentage):
        return self._java_ref.quantile(float(percentage))

    def cdf(self, val):
        return self._java_ref.cdf(float(val))

    def sum(self, val, inclusive=True):
        return self._java_ref.sum(float(val), inclusive)

    def percent_population_over_range(self, start, stop):
        return self._java_ref.percentPopulationOverRange(float(start), float(stop))

    def total_sample_size(self):
        return self._java_ref.totalSampleSize()

    def count(self, bins):
        return PrimitiveLongArrayType().from_java(self._java_ref.count(bins))

    def get_total_count(self):
        return self._java_ref.getTotalCount()

    def get_num_bins(self):
        return self._java_ref.getNumBins()

    def get_min_value(self):
        return self._java_ref.getMinValue()

    def get_max_value(self):
        return self._java_ref.getMaxValue()



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from .query import Query


class AggregationQuery(Query):
    """
    Base aggregation query class that wraps GeoWave aggregation queries.
    """

    def __init__(self, java_ref, java_transformer):
        super().__init__(java_ref, java_transformer)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.config import geowave_pkg
from ..statistic import FieldStatistic
from ..statistic_type import FieldStatisticType
from ...base import GeoWaveObject
from ...base.java_transformer import JavaTransformer


class HyperLogLogStatistic(FieldStatistic):
    """
    Provides an estimated cardinality of the number of unique values for an attribute.
    """
    STATS_TYPE = FieldStatisticType(geowave_pkg.core.store.statistics.field.HyperLogLogStatistic.STATS_TYPE)

    def __init__(self, type_name=None, field_name=None, precision=16, java_ref=None):
        if java_ref is None:
            if type_name is None and field_name is None:
                java_ref = geowave_pkg.core.store.statistics.field.HyperLogLogStatistic()
            else:
                java_ref = geowave_pkg.core.store.statistics.field.HyperLogLogStatistic(type_name, field_name, precision)
        super().__init__(java_ref, HyperLogLogTransformer())

    def set_precision(self, precision):
        self._java_ref.setPrecision(precision)

    def get_precision(self):
        return self._java_ref.getPrecision()


class HyperLogLogTransformer(JavaTransformer):
    def transform(self, j_object):
        return HyperLogLogPlus(j_object)


class HyperLogLogPlus(GeoWaveObject):
    def cardinality(self):
        return self._java_ref.cardinality()



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from datetime import datetime

from shapely.geometry import Polygon

from pygw.index import SpatialIndexBuilder
from pygw.query import VectorQueryBuilder
from pygw.query import FilterFactory

from .conftest import POINT_TYPE_ADAPTER
from .conftest import write_test_data
from .conftest import results_as_list


def test_cql_query(test_ds):
    # given
    index = SpatialIndexBuilder().set_name("idx1").create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)
    write_test_data(test_ds, index)

    # when
    qbldr = VectorQueryBuilder()
    constraints_factory = qbldr.constraints_factory()
    # filter encompasses 10 features (1, 1) - (10, 10)
    constraints = constraints_factory.cql_constraints("BBOX(the_geom, 0.5, 0.5, 10.5, 10.5)")
    qbldr.constraints(constraints)
    res = results_as_list(test_ds.query(qbldr.build()))

    # then
    assert len(res) == 10


def test_query_spatial(test_ds):
    # given
    index = SpatialIndexBuilder().set_name("idx1").create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)
    write_test_data(test_ds, index)

    # when
    qbldr = VectorQueryBuilder()
    stcb = qbldr.constraints_factory().spatial_temporal_constraints()
    # polygon encompasses 10 features (1, 1) - (10, 10)
    stcb.spatial_constraints(Polygon([[0.5, 0.5], [0.5, 10.5], [10.5, 10.5], [10.5, 0.5], [0.5, 0.5]]))
    stcb.spatial_constraints_compare_operation("CONTAINS")
    qbldr.constraints(stcb.build())
    res = results_as_list(test_ds.query(qbldr.build()))

    # then
    assert len(res) == 10


def test_query_temporal(test_ds):
    # given
    index = SpatialIndexBuilder().set_name("idx1").create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)
    write_test_data(test_ds, index)

    # when
    qbldr = VectorQueryBuilder()
    stcb = qbldr.constraints_factory().spatial_temporal_constraints()
    # time range encompasses 10 features (1, 1) - (10, 10)
    stcb.add_time_range(datetime.utcfromtimestamp(1), datetime.utcfromtimestamp(11))
    qbldr.constraints(stcb.build())
    res = results_as_list(test_ds.query(qbldr.build()))

    for feature in res:
        print(feature.get_id())
        print(feature.get_default_geometry())
    # then
    assert len(res) == 10


def test_query_spatial_temporal(test_ds):
    # given
    index = SpatialIndexBuilder().set_name("idx1").create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)
    write_test_data(test_ds, index)

    # when
    qbldr = VectorQueryBuilder()
    stcb = qbldr.constraints_factory().spatial_temporal_constraints()
    # polygon encompasses 10 features (1, 1) - (10, 10)
    stcb.spatial_constraints(Polygon([[0.5, 0.5], [0.5, 10.5], [10.5, 10.5], [10.5, 0.5], [0.5, 0.5]]))
    stcb.spatial_constraints_compare_operation("CONTAINS")
    # time range encompasses 10 features (5, 5) - (14, 14)
    stcb.add_time_range(datetime.utcfromtimestamp(5), datetime.utcfromtimestamp(15))
    qbldr.constraints(stcb.build())
    res = results_as_list(test_ds.query(qbldr.build()))

    # then
    assert len(res) == 6


def test_query_filter(test_ds):
    # given
    index = SpatialIndexBuilder().set_name("idx1").create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)
    write_test_data(test_ds, index)

    # when
    qbldr = VectorQueryBuilder()
    filter_factory = FilterFactory()
    # filter encompasses 10 features (1, 1) - (10, 10)
    bbox_filter = filter_factory.bbox(filter_factory.property("the_geom"), 0.5, 0.5, 10.5, 10.5, "EPSG:4326")
    constraints = qbldr.constraints_factory().filter_constraints(bbox_filter)
    qbldr.constraints(constraints)
    res = results_as_list(test_ds.query(qbldr.build()))

    # then
    assert len(res) == 10



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.config import geowave_pkg, java_gateway
from pygw.store import DataStoreOptions
from py4j.java_collections import MapConverter


class CassandraOptions(DataStoreOptions):
    """
    Cassandra data store options.
    """

    def __init__(self):
        super().__init__(geowave_pkg.datastore.cassandra.config.CassandraRequiredOptions())

    def set_contact_points(self, contact_points):
        """
        Sets a single contact point or a comma delimited set of contact points to
        connect to the Cassandra cluster.

        Args:
            contact_point (str): The contact point(s) to connect to.
        """
        self._java_ref.setContactPoints(contact_points)

    def get_contact_points(self):
        """
        Returns:
            The contact points of the Cassandra cluster.
        """
        return self._java_ref.getContactPoints()

    def set_datacenter(self, datacenter):
        """
        Sets the local datacenter.

        Args:
            datacenter (str): The datacenter to connect to.
        """
        self._java_ref.setDatacenter(datacenter)

    def get_datacenter(self):
        """
        Returns:
            The local datacenter of the Cassandra cluster.
        """
        return self._java_ref.getDatacenter()

    def set_batch_write_size(self, batch_write_size):
        """
        Sets the number of entries to be gathered before performing a batch write
        operation on the data store.

        Args:
            batch_write_size (int): The number of entries to write in batch write operations.
        """
        self._base_options.setBatchWriteSize(batch_write_size)

    def get_batch_write_size(self):
        """
        Returns:
            The number of entries to write in batch write operations.
        """
        return self._base_options.getBatchWriteSize()

    def set_durable_writes(self, durable_writes):
        """
        Sets whether or not to write to commit log for durability, configured only
        on creation of new keyspace.

        Args:
            durable_writes (bool): Whether or not to enable durable writes.
        """
        self._base_options.setDurableWrites(durable_writes)

    def is_durable_writes(self):
        """
        Returns:
            True if durable writes are enabled, False otherwise.
        """
        return self._base_options.isDurableWrites()

    def set_replication_factor(self, replication_factor):
        """
        Sets the number of replicas to use when creating a new keyspace.

        Args:
            replication_factor (int): The number of replicas.
        """
        self._base_options.setReplicationFactor(replication_factor)

    def get_replication_factor(self):
        """
        Returns:
            The number of replicas to use when creating a new keyspace.
        """
        return self._base_options.getReplicationFactor()

    def set_gc_grace_seconds(self, gc_grace_seconds):
        """
        Sets the gc_grace_seconds for each table created. Defaults to 10 days and major
        compaction should be triggered at least as often.

        Args:
            gc_grace_seconds (int): The gc_grace_seconds to set on the table.
        """
        self._base_options.setGcGraceSeconds(gc_grace_seconds)

    def get_gc_grace_seconds(self):
        """
        Returns:
            The gc_grace_seconds applied to new tables.
        """
        return self._base_options.getGcGraceSeconds()

    def set_table_options(self, table_options):
        """
        Sets additional table options for each new table created.

        Args:
            table_options (dictionary): The table options to apply to each new table.
        """
        self._base_options.setTableOptions(MapConverter().convert(table_options, java_gateway._gateway_client))

    def get_table_options(self):
        """
        Returns:
            The table options that are applied to each new table.
        """
        return self._base_options.getTableOptions()

    def set_compaction_strategy(self, compaction_strategy):
        """
        Set the compaction strategy applied to each new Cassandra table. Available options
        are LeveledCompactionStrategy, SizeTieredCompactionStrategy, or TimeWindowCompactionStrategy.

        Args:
            compaction_strategy (str): The compaction strategy to apply to each new table. Available
            options are LeveledCompactionStrategy, SizeTieredCompactionStrategy, or TimeWindowCompactionStrategy.
        """
        self._base_options.setCompactionStrategyStr(compaction_strategy)

    def get_compaction_strategy(self):
        """
        Returns:
            The compaction strategy applied to each new Cassandra table.
        """
        return self._base_options.getCompactionStrategyStr()


#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

class JavaTransformer:
    """
    Base class for transforming Java objects to an appropriate Python counterpart.  All extending classes should
    overwrite the `transform` method to perform the transformation.
    """

    def transform(self, j_object):
        """
        Transforms a Java object into a Python-friendly variant.

        Args:
            j_object (java object): The java object to transform.
        Raises:
            NotImplementedError: This is a base class and not intended to be used directly.
        Returns:
            A Python-friendly equivalent of the query result.
        """
        raise NotImplementedError


class NoOpTransformer(JavaTransformer):
    """
    Transformer that passes through the Java object.
    """

    def transform(self, j_object):
        """
        Pass through the given java object.

        Args:
            j_object (Java Object): An Java object.
        Returns:
            The Java object.
        """
        return j_object



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.config import geowave_pkg
from ..statistic import FieldStatistic
from ..statistic_type import FieldStatisticType
from ...base import GeoWaveObject
from ...base.java_transformer import JavaTransformer


class NumericStatsStatistic(FieldStatistic):
    """
    Tracks the min, max, count, mean, sum, variance and standard deviation of a numeric attribute.
    """
    STATS_TYPE = FieldStatisticType(geowave_pkg.core.store.statistics.field.NumericStatsStatistic.STATS_TYPE)

    def __init__(self, type_name=None, field_name=None, java_ref=None):
        if java_ref is None:
            if type_name is None and field_name is None:
                java_ref = geowave_pkg.core.store.statistics.field.NumericStatsStatistic()
            else:
                java_ref = geowave_pkg.core.store.statistics.field.NumericStatsStatistic(type_name, field_name)
        super().__init__(java_ref, StatsTransformer())


class StatsTransformer(JavaTransformer):
    def transform(self, j_object):
        return Stats(j_object)


class Stats(GeoWaveObject):
    def count(self):
        return self._java_ref.count()

    def mean(self):
        return self._java_ref.mean()

    def sum(self):
        return self._java_ref.sum()

    def population_variance(self):
        return self._java_ref.populationVariance()

    def population_standard_deviation(self):
        return self._java_ref.populationStandardDeviation()

    def sample_variance(self):
        return self._java_ref.sampleVariance()

    def sample_standard_deviation(self):
        return self._java_ref.sampleStandardDeviation()

    def min(self):
        return self._java_ref.min()

    def max(self):
        return self._java_ref.max()



# Packages up pygw so it's pip-installable
from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()


def get_version():
    try:
        from maven_version import get_maven_version
        version = get_maven_version()
    except ModuleNotFoundError:
        # If maven version isn't found, it must be from the distribution
        from pkg_resources import get_distribution
        from pkg_resources import DistributionNotFound
        version = get_distribution('pygw').version
    return version


setup(
    name='pygw',
    author='GeoWave Contributors',
    author_email='geowave.python@gmail.com',
    description='GeoWave bindings for Python3',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://locationtech.github.io/geowave/',
    project_urls={
        'Documentation': 'https://locationtech.github.io/geowave/pydocs/',
        'Source': 'https://github.com/locationtech/geowave/tree/master/python/src/main/python',
    },
    version=get_version(),
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    install_requires=['py4j==0.10.8.1', 'shapely==1.7'],
    python_requires='>=3,<3.8'  # py4j does not support python 3.8 yet
)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
"""
The `gateway` module includes a singleton object of type GatewayConfiguration called `gateway_config` that handles all
communication between python and the Py4J Java Gateway.

By default, the configuration will attempt to connect to a Py4J Java Gateway running on the local machine.  To use a
gateway on a separate machine, the address and port can be specified with the `PYGW_GATEWAY_ADDRESS` and
`PYGW_GATEWAY_PORT` environment variables.  Alternatively, this module can be used to provide pygw with a Py4J Java
Gateway that has already been initialized by importing the `gateway_config` from this module and calling `set_gateway`
with the gateway object prior to importing any other pygw classes.  It is important to note that GeoWave must be on the
classpath of the gateway that is being provided.

The following snippet shows how a custom gateway (such as one from pyspark) can be provided:
```python
# This snippet should only be called once at the beginning of the application lifecycle
from pygw.gateway import gateway_config
gateway_config.set_gateway(sparkContext._gateway)

# Continue to use pygw normally
```

NOTE: the GatewayConfiguration has an `init()` method. This is INTENTIONALLY not an `__init__` method. Initialization is
attempted when the configuration is imported.
"""
from os import environ

from py4j.java_gateway import JavaGateway
from py4j.java_gateway import GatewayParameters
from py4j.java_gateway import DEFAULT_ADDRESS
from py4j.java_gateway import DEFAULT_PORT

from py4j.protocol import Py4JNetworkError

PYGW_ADDRESS_ENV = 'PYGW_GATEWAY_ADDRESS'
PYGW_PORT_ENV = 'PYGW_GATEWAY_PORT'


class GatewayConfiguration:
    """
    This class sets up communication between Python and the GeoWave logic running
    on a JVM.
    """

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(GatewayConfiguration, cls).__new__(cls)
        return cls.instance

    def __init__(self):
        self.is_initialized = False
        self.GATEWAY = None
        self.geowave_version = None

    def set_gateway(self, gateway):
        if self.GATEWAY is not None:
            raise GatewayConfiguration.PyGwJavaGatewayAlreadyInitializedError(
                "The pygw gateway has already been initialized. Set the gateway before importing any other pygw "
                "classes.")
        self.GATEWAY = gateway

    def init(self):
        """
        Sets up the Py4J Gateway.  This is called automatically when other pygw classes are imported.
        """
        if not self.is_initialized:
            try:
                if self.GATEWAY is None:
                    gateway_address = environ.get(PYGW_ADDRESS_ENV, DEFAULT_ADDRESS)
                    gateway_port = environ.get(PYGW_PORT_ENV, DEFAULT_PORT)
                    self.GATEWAY = JavaGateway(
                        gateway_parameters=GatewayParameters(auto_field=True, address=gateway_address,
                                                             port=gateway_port))
                    try:
                        self.geowave_version = self.GATEWAY.jvm.org.locationtech.geowave.core.cli.VersionUtils\
                            .getVersion()
                    except TypeError as e:
                        raise GatewayConfiguration.PyGwGeoWaveNotFoundInGateway(
                            "GeoWave was not found in the configured gateway.  Make sure GeoWave jars are available "
                            "on the classpath of the running gateway.") from e

                    self.is_initialized = True

            except Py4JNetworkError as e:
                raise GatewayConfiguration.PyGwJavaGatewayNotStartedError(
                    "The GeoWave Py4J Java Gateway must be running before you can use pygw.") from e

    class PyGwGeoWaveNotFoundInGateway(Exception):
        pass

    class PyGwJavaGatewayNotStartedError(Exception):
        pass

    class PyGwJavaGatewayAlreadyInitializedError(Exception):
        pass


gateway_config = GatewayConfiguration()

__all__ = ["gateway_config"]



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.query import Query


class StatisticQuery(Query):
    """
    Base statistic query class that wraps GeoWave statistic queries.
    """

    def __init__(self, java_ref, result_transformer):
        super().__init__(java_ref, result_transformer)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.config import geowave_pkg
from ..statistic import IndexStatistic
from ..statistic_type import IndexStatisticType
from ...base.java_transformer import JavaTransformer
from ...base.type_conversions import PrimitiveByteArrayType


class FieldVisibilityCountStatistic(IndexStatistic):
    """
    Maintains a count of entries for every visibility.
    """
    STATS_TYPE = IndexStatisticType(geowave_pkg.core.store.statistics.index.FieldVisibilityCountStatistic.STATS_TYPE)

    def __init__(self, index_name=None, java_ref=None):
        if java_ref is None:
            if index_name is None:
                java_ref = geowave_pkg.core.store.statistics.index.FieldVisibilityCountStatistic()
            else:
                java_ref = geowave_pkg.core.store.statistics.index.FieldVisibilityCountStatistic(index_name)
        super().__init__(java_ref, FieldVisibilityCountTransformer())


class FieldVisibilityCountTransformer(JavaTransformer):
    def transform(self, j_object):
        field_visibilities = {}
        byte_array_type = PrimitiveByteArrayType()
        for key, value in j_object.items():
            field_visibilities[byte_array_type.from_java(key.getBytes())] = value
        return field_visibilities



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.config import geowave_pkg
from .statistic_binning_strategy import StatisticBinningStrategy
from .binning_strategy.composite_binning_strategy import CompositeBinningStrategy
from .binning_strategy.data_type_binning_strategy import DataTypeBinningStrategy
from .binning_strategy.field_value_binning_strategy import FieldValueBinningStrategy
from .binning_strategy.numeric_range_field_value_binning_strategy import NumericRangeFieldValueBinningStrategy
from .binning_strategy.partition_binning_strategy import PartitionBinningStrategy
from .binning_strategy.spatial_field_value_binning_strategy import SpatialFieldValueBinningStrategy
from .binning_strategy.time_range_field_value_binning_strategy import TimeRangeFieldValueBinningStrategy

__binning_strategy_mappings = {
    geowave_pkg.core.store.statistics.binning.CompositeBinningStrategy.NAME:
        CompositeBinningStrategy,
    geowave_pkg.core.store.statistics.binning.DataTypeBinningStrategy.NAME:
        DataTypeBinningStrategy,
    geowave_pkg.core.store.statistics.binning.PartitionBinningStrategy.NAME:
        PartitionBinningStrategy,
    geowave_pkg.core.store.statistics.binning.NumericRangeFieldValueBinningStrategy.NAME:
        NumericRangeFieldValueBinningStrategy,
    geowave_pkg.core.geotime.store.statistics.binning.SpatialFieldValueBinningStrategy.NAME:
        SpatialFieldValueBinningStrategy,
    geowave_pkg.core.geotime.store.statistics.binning.TimeRangeFieldValueBinningStrategy.NAME:
        TimeRangeFieldValueBinningStrategy,
    geowave_pkg.core.store.statistics.binning.FieldValueBinningStrategy.NAME:
        FieldValueBinningStrategy,
}


def map_binning_strategy(j_binning_strategy):
    if j_binning_strategy is None:
        return None
    strategy_name = j_binning_strategy.getStrategyName()
    if strategy_name in __binning_strategy_mappings:
        return __binning_strategy_mappings[strategy_name](java_ref=j_binning_strategy)
    return StatisticBinningStrategy(j_binning_strategy)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from datetime import datetime

from pygw.index import SpatialIndexBuilder
from pygw.query import VectorAggregationQueryBuilder

from .conftest import POINT_TYPE_ADAPTER, POINT_GEOMETRY_FIELD, POINT_TIME_FIELD, POINT_TYPE_NAME, POINT_NUMBER_FIELD
from .conftest import write_test_data
from ..base import Envelope, Interval


def setup_query_builder(test_ds):
    # given
    index = SpatialIndexBuilder().set_name("idx1").create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)
    write_test_data(test_ds, index)

    qbldr = VectorAggregationQueryBuilder()
    constraints_factory = qbldr.constraints_factory()
    # filter encompasses 10 features (1, 1) - (10, 10)
    constraints = constraints_factory.cql_constraints("BBOX(the_geom, 0.5, 0.5, 10.5, 10.5)")
    qbldr.constraints(constraints)
    return qbldr


def test_bbox_aggregation(test_ds):
    # given
    qbldr = setup_query_builder(test_ds)

    # when
    qbldr.bbox_of_results(POINT_TYPE_NAME)
    res = test_ds.aggregate(qbldr.build())

    # then
    assert isinstance(res, Envelope)
    assert res.get_min_x() == 1.0
    assert res.get_min_y() == 1.0
    assert res.get_max_x() == 10.0
    assert res.get_max_y() == 10.0

    # when
    qbldr.bbox_of_results_for_geometry_field(POINT_TYPE_NAME, POINT_GEOMETRY_FIELD)
    res = test_ds.aggregate(qbldr.build())

    # then
    assert isinstance(res, Envelope)
    assert res.get_min_x() == 1.0
    assert res.get_min_y() == 1.0
    assert res.get_max_x() == 10.0
    assert res.get_max_y() == 10.0


def test_time_range_aggregation(test_ds):
    # given
    qbldr = setup_query_builder(test_ds)

    # when
    qbldr.time_range_of_results(POINT_TYPE_NAME)
    res = test_ds.aggregate(qbldr.build())

    # then
    assert isinstance(res, Interval)
    assert res.get_start() == datetime.utcfromtimestamp(1)  # Start Date
    assert res.get_end() == datetime.utcfromtimestamp(10)  # End Date

    # when
    qbldr.time_range_of_results_for_time_field(POINT_TYPE_NAME, POINT_TIME_FIELD)
    res = test_ds.aggregate(qbldr.build())

    # then
    assert isinstance(res, Interval)
    assert res.get_start() == datetime.utcfromtimestamp(1)  # Start Date
    assert res.get_end() == datetime.utcfromtimestamp(10)  # End Date


def test_math_aggregations(test_ds):
    # given
    qbldr = setup_query_builder(test_ds)

    # when
    qbldr.max(POINT_TYPE_NAME, POINT_NUMBER_FIELD)
    res = test_ds.aggregate(qbldr.build())

    # then
    assert res == 10.0  # Maximum number

    # when
    qbldr.min(POINT_TYPE_NAME, POINT_NUMBER_FIELD)
    res = test_ds.aggregate(qbldr.build())

    # then
    assert res == 1.0  # Minimum number

    # when
    qbldr.sum(POINT_TYPE_NAME, POINT_NUMBER_FIELD)
    res = test_ds.aggregate(qbldr.build())

    # then
    assert res == 55.0  # 1 + 2 + 3 + 4 + 5 + 6 + 7 + 8 + 9 + 10


def test_count_aggregation(test_ds):
    # given
    qbldr = setup_query_builder(test_ds)

    # when
    qbldr.count(POINT_TYPE_NAME)
    res = test_ds.aggregate(qbldr.build())

    # then
    assert res == 10



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from datetime import datetime

from shapely.geometry.base import BaseGeometry

from pygw.base import GeoWaveObject
from pygw.base.type_conversions import GeometryType
from pygw.config import java_gateway
from pygw.config import java_pkg
from pygw.config import reflection_util


def _j_match_action(match_action):
    return java_pkg.org.opengis.filter.MultiValuedFilter.MatchAction.valueOf(match_action.upper())


# These functions are needed in order to invoke java methods that are named with
# reserved python keywords such as and, or, and not
def _invoke_filter_list_method_by_name(j_filter_factory, name, filters):
    filter_factory_class = j_filter_factory.getClass()
    list_class = reflection_util.classForName("java.util.List")
    class_array = java_gateway.new_array(java_pkg.java.lang.Class, 1)
    class_array[0] = list_class
    method = filter_factory_class.getMethod(name, class_array)
    filter_list = java_pkg.java.util.ArrayList()
    for filter_object in filters:
        filter_list.append(filter_object)
    objects_array = java_gateway.new_array(java_pkg.java.lang.Object, 1)
    objects_array[0] = filter_list
    return method.invoke(j_filter_factory, objects_array)


def _invoke_filter_method_by_name(j_filter_factory, name, filter_object):
    filter_factory_class = j_filter_factory.getClass()
    filter_class = reflection_util.classForName("org.opengis.filter.Filter")
    class_array = java_gateway.new_array(java_pkg.java.lang.Class, 1)
    class_array[0] = filter_class
    method = filter_factory_class.getMethod(name, class_array)
    objects_array = java_gateway.new_array(java_pkg.java.lang.Object, 1)
    objects_array[0] = filter_object
    return method.invoke(j_filter_factory, objects_array)


class FilterFactory(GeoWaveObject):
    """
    Filter factory for constructing filters to be used in vector queries. Methods
    of this factory generally return either a Filter or Expression which can be used
    in additional method calls.
    """

    def __init__(self):
        j_filter_factory = java_pkg.org.geotools.filter.FilterFactoryImpl()
        super().__init__(j_filter_factory)

    def id(self, fids):
        """
        Constructs a filter that matches a set of feature IDs.

        Args:
            fids (list of str): The list of feature IDs to match.
        Returns:
            A Filter with the given feature IDs.
        """
        j_fids = java_gateway.new_array(java_pkg.org.opengis.filter.identity.FeatureId, len(fids))
        for idx, fid in enumerate(fids):
            if isinstance(fid, str):
                j_fids[idx] = self.feature_id(fid)
            else:
                j_fids[idx] = fid
        return self._java_ref.id(j_fids)

    def feature_id(self, fid):
        """
        Constructs a filter that matches a specific feature ID.

        Args:
            fid (str): The feature ID.
        Returns:
            A Filter with the given feature ID.
        """
        return self._java_ref.featureId(fid)

    def gml_object_id(self, object_id):
        """
        Constructs a filter that matches a specific gml object ID.

        Args:
            object_id (str): The gml object ID.
        Returns:
            A Filter with the given gml object ID.
        """
        return self._java_ref.gmlObjectId(object_id)

    def property(self, name):
        """
        Constructs an expression that references the given property name.

        Args:
            name (str): The property name.
        Returns:
            An Expression with the given property name.
        """
        return self._java_ref.property(name)

    def literal(self, value):
        """
        Constructs an expression with the given literal value.

        Args:
            value (any): The literal value to use.
        Returns:
            An Expression with the given literal value.
        """
        if isinstance(value, datetime):
            # Convert the date to a string
            value = value.strftime("%Y-%m-%dT%H:%M:%S")
        if isinstance(value, str):
            # Prevent Py4J from assuming the string matches up with the char variant method
            filter_factory_class = self._java_ref.getClass()
            object_class = reflection_util.classForName("java.lang.Object")
            class_array = java_gateway.new_array(java_pkg.java.lang.Class, 1)
            class_array[0] = object_class
            method = filter_factory_class.getMethod("literal", class_array)
            objects_array = java_gateway.new_array(java_pkg.java.lang.Object, 1)
            objects_array[0] = value
            return method.invoke(self._java_ref, objects_array)
        if isinstance(value, BaseGeometry):
            return self._java_ref.literal(GeometryType().to_java(value))
        return self._java_ref.literal(value)

    def add(self, expr1, expr2):
        """
        Constructs an expression which adds two other expressions.

        Args:
            expr1 (Expression): The first expression.
            expr2 (Expression): The second expression.
        Returns:
            An Expression which represents [expr1 + expr2].
        """
        return self._java_ref.add(expr1, expr2)

    def subtract(self, expr1, expr2):
        """
        Constructs an expression which subtracts one expression from another.

        Args:
            expr1 (Expression): The expression to subtract from.
            expr2 (Expression): The expression to subtract.
        Returns:
            An Expression which represents [expr1 - expr2].
        """
        return self._java_ref.subtract(expr1, expr2)

    def multiply(self, expr1, expr2):
        """
        Constructs an expression which multiplies two other expressions.

        Args:
            expr1 (Expression): The first expression.
            expr2 (Expression): The second expression.
        Returns:
            An Expression which represents [expr1 * expr2].
        """
        return self._java_ref.multiply(expr1, expr2)

    def divide(self, expr1, expr2):
        """
        Constructs an expression which divides one expression by another.

        Args:
            expr1 (Expression): The expression to divide.
            expr2 (Expression): The expression to divide by.
        Returns:
            An Expression which represents [expr1 / expr2].
        """
        return self._java_ref.divide(expr1, expr2)

    def function(self, name, expressions):
        """
        Constructs an expression by passing a set of expressions to an expression function.

        Args:
            name (str): The name of the function.
            expressions (list of Expression): The expressions to use in the function.
        Returns:
            An Expression which represents the result of the function.
        """
        j_expressions = java_gateway.new_array(java_pkg.org.opengis.filter.expression.Expression, len(expressions))
        for idx, expression in enumerate(expressions):
            j_expressions[idx] = expression
        return self._java_ref.function(name, j_expressions)

    def and_(self, filters):
        """
        Constructs a filter which passes when all given filters pass.

        Args:
            filters (list of Filter): The filters to check.
        Returns:
            A Filter that passes when all given Filters pass.
        """
        return _invoke_filter_list_method_by_name(self._java_ref, "and", filters)

    def or_(self, filters):
        """
        Constructs a filter which passes when any of the given filters pass.

        Args:
            filters (list of Filter): The filters to check.
        Returns:
            A Filter that passes when one of the given Filters pass.
        """
        return _invoke_filter_list_method_by_name(self._java_ref, "or", filters)

    def not_(self, filter_to_check):
        """
        Constructs a filter that passes when the given filter does NOT pass.

        Args:
            filter_to_check (Filter): The filter to check.
        Returns:
            A Filter that passes when the given filter does NOT pass.
        """
        return _invoke_filter_method_by_name(self._java_ref, "not", filter_to_check)

    def between(self, expr, lower, upper, match_action=None):
        """
        Constructs a filter that passes when the given expression falls between a
        lower and upper expression.

        Args:
            expr (Expression): The expression to check.
            lower (Expression): The lower bound.
            upper (Expression): The upper bound.
            match_action (str): The match action to use. Default is 'ANY'.
        Returns:
            A Filter that passes when the given expression falls between a
            lower and upper expression.
        """
        if match_action is None:
            return self._java_ref.between(expr, lower, upper)
        else:
            return self._java_ref.between(expr, lower, upper, _j_match_action(match_action))

    def equals(self, expr1, expr2):
        """
        Constructs a filter that passes when the given expressions are equal.

        Args:
            expr1 (Expression): The first expression.
            expr2 (Expression): The second expression.
        Returns:
            A Filter that passes when the given expressions are equal.
        """
        return self._java_ref.equals(expr1, expr2)

    def equal(self, expr1, expr2, match_case, match_action=None):
        """
        Constructs a filter that passes when the given expressions are equal.

        Args:
            expr1 (Expression): The first expression.
            expr2 (Expression): The second expression.
            match_case (bool): Whether or not to match case with strings.
            match_action (str): The match action to use. Default is 'ANY'.
        Returns:
            A Filter that passes when the given expressions are equal.
        """
        if match_action is None:
            return self._java_ref.equal(expr1, expr2, match_case)
        else:
            return self._java_ref.equal(expr1, expr2, match_case, _j_match_action(match_action))

    def not_equals(self, expr1, expr2):
        """
        Constructs a filter that passes when the given expressions are NOT equal.

        Args:
            expr1 (Expression): The first expression.
            expr2 (Expression): The second expression.
        Returns:
            A Filter that passes when the given expressions are NOT equal.
        """
        return self._java_ref.notEqual(expr1, expr2)

    def not_equal(self, expr1, expr2, match_case, match_action=None):
        """
        Constructs a filter that passes when the given expressions are NOT equal.

        Args:
            expr1 (Expression): The first expression.
            expr2 (Expression): The second expression.
            match_case (bool): Whether or not to match case with strings.
            match_action (str): The match action to use. Default is 'ANY'.
        Returns:
            A Filter that passes when the given expressions are NOT equal.
        """
        if match_action is None:
            return self._java_ref.notEqual(expr1, expr2, match_case)
        else:
            return self._java_ref.notEqual(expr1, expr2, match_case, _j_match_action(match_action))

    def greater(self, expr1, expr2, match_case=None, match_action=None):
        """
        Constructs a filter that passes when the first expression is greater than
        the second.

        Args:
            expr1 (Expression): The first expression.
            expr2 (Expression): The second expression.
            match_case (bool): Whether or not to match case with strings. Default is None.
            match_action (str): The match action to use. Default is 'ANY'.
        Returns:
            A Filter that passes when the first expression is greater than the
            second.
        """
        if match_case is None:
            return self._java_ref.greater(expr1, expr2)
        elif match_action is None:
            return self._java_ref.greater(expr1, expr2, match_case)
        else:
            return self._java_ref.greater(expr1, expr2, match_case, _j_match_action(match_action))

    def greater_or_equal(self, expr1, expr2, match_case=None, match_action=None):
        """
        Constructs a filter that passes when the first expression is greater than
        or equal to the second.

        Args:
            expr1 (Expression): The first expression.
            expr2 (Expression): The second expression.
            match_case (bool): Whether or not to match case with strings. Default is None.
            match_action (str): The match action to use. Default is 'ANY'.
        Returns:
            A Filter that passes when the first expression is greater than or equal
            to the second.
        """
        if match_case is None:
            return self._java_ref.greaterOrEqual(expr1, expr2)
        elif match_action is None:
            return self._java_ref.greaterOrEqual(expr1, expr2, match_case)
        else:
            return self._java_ref.greaterOrEqual(expr1, expr2, match_case, _j_match_action(match_action))

    def less(self, expr1, expr2, match_case=None, match_action=None):
        """
        Constructs a filter that passes when the first expression is less than
        the second.

        Args:
            expr1 (Expression): The first expression.
            expr2 (Expression): The second expression.
            match_case (bool): Whether or not to match case with strings. Default is None.
            match_action (str): The match action to use. Default is 'ANY'.
        Returns:
            A Filter that passes when the first expression is less than the
            second.
        """
        if match_case is None:
            return self._java_ref.less(expr1, expr2)
        elif match_action is None:
            return self._java_ref.less(expr1, expr2, match_case)
        else:
            return self._java_ref.less(expr1, expr2, match_case, _j_match_action(match_action))

    def less_or_equal(self, expr1, expr2, match_case=None, match_action=None):
        """
        Constructs a filter that passes when the first expression is les than
        or equal to the second.

        Args:
            expr1 (Expression): The first expression.
            expr2 (Expression): The second expression.
            match_case (bool): Whether or not to match case with strings. Default is None.
            match_action (str): The match action to use. Default is 'ANY'.
        Returns:
            A Filter that passes when the first expression is less than or equal
            to the second.
        """
        if match_case is None:
            return self._java_ref.lessOrEqual(expr1, expr2)
        elif match_action is None:
            return self._java_ref.lessOrEqual(expr1, expr2, match_case)
        else:
            return self._java_ref.lessOrEqual(expr1, expr2, match_case, _j_match_action(match_action))

    def like(self, expr, pattern, wildcard=None, single_char=None, escape=None, match_case=None, match_action=None):
        """
        Constructs a filter with character string comparison operator with pattern
        matching and specified wildcards.

        Args:
            expr (Expression): The expression to use.
            pattern (str): The pattern to match.
            wildcard (str): The string to use to match any characters.  Default is None.
            single_char (str): The string to use to match a single character.  Default is None.
            escape (str): The string to use to escape a wildcard.  Default is None.
            match_case (bool): Whether or not to match case with strings. Default is None.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when the first expression is greater than the
            second.
        """
        if wildcard is None:
            return self._java_ref.like(expr, pattern)
        elif match_case is None:
            return self._java_ref.like(expr, pattern, wildcard, single_char, escape)
        elif match_action is None:
            return self._java_ref.like(expr, pattern, wildcard, single_char, escape, match_case)
        else:
            return self._java_ref.like(expr, pattern, wildcard, single_char, escape, match_case,
                                       _j_match_action(match_action))

    def is_null(self, expr):
        """
        Constructs a filter that passes when the given expression is null.

        Args:
            expr (Expression): The expression to check.
        Returns:
            A Filter that passes when the given epxression is null.
        """
        return self._java_ref.isNull(expr)

    def bbox(self, geometry_expr, minx, miny, maxx, maxy, srs, match_action=None):
        """
        Constructs a filter that passes when the given geometry expression is within
        the given bounding box.

        Args:
            geometry_expr (Expression): An expression which represents a geometry.
            minx (float): The minimum X value of the bounding box.
            miny (float): The minimum Y value of the bounding box.
            maxx (float): The maximum X value of the bounding box.
            maxy (float): The maximum Y value of the bounding box.
            srs (str): The spatial reference system of the geometry.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when the given geometry is within the bounding box.
        """
        if match_action is None:
            return self._java_ref.bbox(geometry_expr, minx * 1.0, miny * 1.0, maxx * 1.0, maxy * 1.0, srs)
        else:
            return self._java_ref.bbox(geometry_expr, minx * 1.0, miny * 1.0, maxx * 1.0, maxy * 1.0, srs,
                                       _j_match_action(match_action))

    def bbox_expr(self, geometry_expr, bbox_expr):
        """
        Constructs a filter that passes when the given geometry expression is within the
        given bounding box expression.

        Args:
            geometry_expr (Expression): An expression which represents a geometry.
            bbox_expr (Expression): An expression which represents a bounding box.
        Returns:
            A Filter that passes when the given geometry is within the bounding box.
        """
        return self._java_ref.bbox(geometry_expr, bbox_expr)

    def beyond(self, geometry_expr1, geometry_expr2, distance, units, match_action=None):
        """
        Constructs a filter that passes when a given geometry is beyond a certain distance from
        a second given geometry.

        Args:
            geometry_expr1 (Expression): An expression which represents a geometry.
            geometry_expr2 (Expression): An expression which represents a geometry.
            distance (float): The distance to use.
            units (str): The distance unit.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when the geometry is beyond the distance from the second
            geometry.
        """
        if match_action is None:
            return self._java_ref.beyond(geometry_expr1, geometry_expr2, distance * 1.0, units)
        else:
            return self._java_ref.beyond(geometry_expr1, geometry_expr2, distance * 1.0, units,
                                         _j_match_action(match_action))

    def contains(self, geometry_expr1, geometry_expr2, match_action=None):
        """
        Constructs a filter that passes when the first geometry expression contains the
        second geometry expression.

        Args:
            geometry_expr1 (Expression): An expression which represents the geometry to check against.
            geometry_expr2 (Expression): An expression which represents the geometry to check.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when the second geometry is contained by the first.
        """
        if match_action is None:
            return self._java_ref.contains(geometry_expr1, geometry_expr2)
        else:
            return self._java_ref.contains(geometry_expr1, geometry_expr2, _j_match_action(match_action))

    def crosses(self, geometry_expr1, geometry_expr2, match_action=None):
        """
        Constructs a filter that passes when a given geometry crosses another given geometry.

        Args:
            geometry_expr1 (Expression): An expression which represents a geometry.
            geometry_expr2 (Expression): An expression which represents a geometry.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given geometry crosses another given geometry.
        """
        if match_action is None:
            return self._java_ref.crosses(geometry_expr1, geometry_expr2)
        else:
            return self._java_ref.crosses(geometry_expr1, geometry_expr2, _j_match_action(match_action))

    def disjoint(self, geometry_expr1, geometry_expr2, match_action=None):
        """
        Constructs a filter that passes when a given geometry is disjoint to another given geometry.

        Args:
            geometry_expr1 (Expression): An expression which represents a geometry.
            geometry_expr2 (Expression): An expression which represents a geometry.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given geometry is disjoint to another given geometry.
        """
        if match_action is None:
            return self._java_ref.disjoint(geometry_expr1, geometry_expr2)
        else:
            return self._java_ref.disjoint(geometry_expr1, geometry_expr2, _j_match_action(match_action))

    def intersects(self, geometry_expr1, geometry_expr2, match_action=None):
        """
        Constructs a filter that passes when a given geometry intersects another given geometry.

        Args:
            geometry_expr1 (Expression): An expression which represents a geometry.
            geometry_expr2 (Expression): An expression which represents a geometry.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given geometry intersects another given geometry.
        """
        if match_action is None:
            return self._java_ref.intersects(geometry_expr1, geometry_expr2)
        else:
            return self._java_ref.intersects(geometry_expr1, geometry_expr2, _j_match_action(match_action))

    def overlaps(self, geometry_expr1, geometry_expr2, match_action=None):
        """
        Constructs a filter that passes when a given geometry overlaps another given geometry.

        Args:
            geometry_expr1 (Expression): An expression which represents a geometry.
            geometry_expr2 (Expression): An expression which represents a geometry.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given geometry overlaps another given geometry.
        """
        if match_action is None:
            return self._java_ref.overlaps(geometry_expr1, geometry_expr2)
        else:
            return self._java_ref.overlaps(geometry_expr1, geometry_expr2, _j_match_action(match_action))

    def touches(self, geometry_expr1, geometry_expr2, match_action=None):
        """
        Constructs a filter that passes when a given geometry touches another given geometry.

        Args:
            geometry_expr1 (Expression): An expression which represents a geometry.
            geometry_expr2 (Expression): An expression which represents a geometry.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given geometry touches another given geometry.
        """
        if match_action is None:
            return self._java_ref.touches(geometry_expr1, geometry_expr2)
        else:
            return self._java_ref.touches(geometry_expr1, geometry_expr2, _j_match_action(match_action))

    def within(self, geometry_expr1, geometry_expr2, match_action=None):
        """
        Constructs a filter that passes when a given geometry is within another given geometry.

        Args:
            geometry_expr1 (Expression): An expression which represents a geometry.
            geometry_expr2 (Expression): An expression which represents a geometry.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given geometry is within another given geometry.
        """
        if match_action is None:
            return self._java_ref.within(geometry_expr1, geometry_expr2)
        else:
            return self._java_ref.within(geometry_expr1, geometry_expr2, _j_match_action(match_action))

    def dwithin(self, geometry_expr1, geometry_expr2, distance, units, match_action=None):
        """
        Constructs a filter that passes when a given geometry is within the specified distance
        of the second geometry.

        Args:
            geometry_expr1 (Expression): An expression which represents a geometry.
            geometry_expr2 (Expression): An expression which represents a geometry.
            distance (float): The distance to use.
            units (str): The unit of distance.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given geometry is within the specified distance of the
            second geometry.
        """
        if match_action is None:
            return self._java_ref.dwithin(geometry_expr1, geometry_expr2, distance * 1.0, units)
        else:
            return self._java_ref.dwithin(geometry_expr1, geometry_expr2, distance * 1.0, units,
                                          _j_match_action(match_action))

    def after(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression occurs after
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression occurs after the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.after(expr1, expr2)
        else:
            return self._java_ref.after(expr1, expr2, _j_match_action(match_action))

    def any_interacts(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression interacts with
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression interacts with the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.anyInteracts(expr1, expr2)
        else:
            return self._java_ref.anyInteracts(expr1, expr2, _j_match_action(match_action))

    def before(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression occurs before
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression occurs before the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.before(expr1, expr2)
        else:
            return self._java_ref.before(expr1, expr2, _j_match_action(match_action))

    def begins(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression begins
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression begins the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.begins(expr1, expr2)
        else:
            return self._java_ref.begins(expr1, expr2, _j_match_action(match_action))

    def begun_by(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression is begun by
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression is begun by the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.begunBy(expr1, expr2)
        else:
            return self._java_ref.begunBy(expr1, expr2, _j_match_action(match_action))

    def during(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression occurs during
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression occurs during the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.during(expr1, expr2)
        else:
            return self._java_ref.during(expr1, expr2, _j_match_action(match_action))

    def ended_by(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression is ended by
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression is ended by the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.endedBy(expr1, expr2)
        else:
            return self._java_ref.endedBy(expr1, expr2, _j_match_action(match_action))

    def ends(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression ends
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression ends the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.ends(expr1, expr2)
        else:
            return self._java_ref.ends(expr1, expr2, _j_match_action(match_action))

    def meets(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression meets
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression meets the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.meets(expr1, expr2)
        else:
            return self._java_ref.meets(expr1, expr2, _j_match_action(match_action))

    def met_by(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression is met by
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression is met by the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.metBy(expr1, expr2)
        else:
            return self._java_ref.metBy(expr1, expr2, _j_match_action(match_action))

    def overlapped_by(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression is overlapped by
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression is overlapped by the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.overlappedBy(expr1, expr2)
        else:
            return self._java_ref.overlappedBy(expr1, expr2, _j_match_action(match_action))

    def tcontains(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression contains
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression contains the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.tcontains(expr1, expr2)
        else:
            return self._java_ref.tcontains(expr1, expr2, _j_match_action(match_action))

    def tequals(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression equals
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression equals the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.tequals(expr1, expr2)
        else:
            return self._java_ref.tequals(expr1, expr2, _j_match_action(match_action))

    def toverlaps(self, expr1, expr2, match_action=None):
        """
        Constructs a filter that passes when a given temporal expression overlaps
        a second temporal expression.

        Args:
            expr1 (Expression): The first temporal expression.
            expr2 (Expression): The second temporal expression.
            match_action (str): The match action to use.  Default is 'ANY'.
        Returns:
            A Filter that passes when a given temporal expression overlaps the
            second temporal expression.
        """
        if match_action is None:
            return self._java_ref.toverlaps(expr1, expr2)
        else:
            return self._java_ref.toverlaps(expr1, expr2, _j_match_action(match_action))



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.base.type_conversions import StringArrayType

from .query import Query
from .base_query_builder import BaseQueryBuilder


class QueryBuilder(BaseQueryBuilder):
    """
    Base query builder for constructing GeoWave queries.  This class should not
    be used directly.  Instead, use one of the derived classes such as
    `pygw.query.vector.vector_query_builder.VectorQueryBuilder`.
    """

    def __init__(self, java_ref, java_transformer):
        super().__init__(java_ref, java_transformer)

    def all_types(self):
        """
        Configure the query to get data from all types. This is the default.

        Returns:
            This query builder.
        """
        self._java_ref.allTypes()
        return self

    def add_type_name(self, type_name):
        """
        Configure the query to get data from a specific type.

        Args:
            type_name (str): The type to get data from.
        Returns:
            This query builder.
        """
        self._java_ref.addTypeName(type_name)
        return self

    def set_type_names(self, type_names):
        """
        Configure the query to get data from a set of types.

        Args:
            type_names (list of str): The types to get data from.
        Returns:
            This query builder.
        """
        self._java_ref.setTypeNames(StringArrayType().to_java(type_names))
        return self

    def subset_fields(self, type_name, field_names):
        """
        Configure the query to get a specific set of fields from a given type.

        Args:
            type_name (str): The type to get from.
            field_names (list of str): The fields to get.
        Returns:
            This query builder.
        """
        self._java_ref.subsetFields(type_name, StringArrayType().to_java(field_names))
        return self

    def all_fields(self):
        """
        Configure the query to get all fields from the given type(s). This is the
        default.

        Returns:
            This query builder.
        """
        self._java_ref.allFields()
        return self

    def build(self):
        """
        Builds the configured query.

        Returns:
            The final constructed `pygw.query.query.Query`.
        """
        return Query(self._java_ref.build(), self._java_transformer)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.config import geowave_pkg
from pygw.store import DataStoreOptions


class RedisOptions(DataStoreOptions):
    """
    Redis data store options.
    """

    def __init__(self):
        super().__init__(geowave_pkg.datastore.redis.config.RedisOptions())

    def set_address(self, address):
        """
        Sets the address of the Redis data store.

        Args:
            address (str): The address of the Redis data store.
        """
        self._java_ref.setAddress(address)

    def get_address(self):
        """
        Returns:
            The address of the Redis data store.
        """
        return self._java_ref.getAddress()

    def set_compression(self, compression):
        """
        Sets the compression to use for the Redis data store. Valid options are `SNAPPY`,
        `L4Z`, or `NONE`.

        Args:
            compression (str): The compressioin to use.
        """
        converter = geowave_pkg.datastore.redis.config.RedisOptions.CompressionConverter()
        self._java_ref.setCompression(converter.convert(compression))

    def get_compression(self):
        """
        Returns:
            The compression used by the data store.
        """
        return self._java_ref.getCompression().name()

    def set_username(self, username):
        """
        Sets the Redis username to be used with Redis AUTH.

        Args:
            username (str): The Redis username to be used with Redis AUTH.
        """
        self._java_ref.setUsername(username)

    def get_username(self):
        """
        Returns:
            The Redis username to be used with Redis AUTH.
        """
        return self._java_ref.getUsername()

    def set_password(self, password):
        """
        Sets the password for the user to be used with Redis AUTH.

        Args:
            password (str): The password for the user to be used with Redis AUTH.
        """
        self._java_ref.setPassword(password)

    def get_password(self):
        """
        Returns:
            The password for the user to be used with Redis AUTH.
        """
        return self._java_ref.getPassword()



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from .geowave_object import GeoWaveObject
from .java_transformer import NoOpTransformer


class CloseableIterator(GeoWaveObject):
    """
    A wrapper for GeoWave CloseableIterators.
    """

    def __init__(self, java_ref, java_transformer=NoOpTransformer()):
        self._java_transformer = java_transformer
        super().__init__(java_ref)

    def __iter__(self):
        return self

    def __next__(self):
        if self._java_ref.hasNext():
            return self._java_transformer.transform(self._java_ref.next())
        else:
            raise StopIteration

    def close(self):
        """
        Closes the Java CloseableIterator.
        """
        self._java_ref.close()



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.base import GeoWaveObject


class Index(GeoWaveObject):
    """
    Wrapper class for a GeoWave Index.
    """

    def get_name(self):
        """
        Returns:
            The name of the index.
        """
        return self._java_ref.getName()

    def get_index_strategy(self):
        """
        Returns:
            The class name of the index strategy of the index.
        """
        j_obj = self._java_ref.getIndexStrategy()
        return j_obj.getClass().toString()

    def get_index_model(self):
        """
        Returns:
            The class name of the index model of the index.
        """
        j_obj = self._java_ref.getIndexModel()
        return j_obj.getClass().toString()



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from py4j.java_gateway import is_instance_of

from pygw.config import java_gateway


class GeoWaveObject:
    """
    Base Class for pygw objects that wrap Java objects.
    """

    def __init__(self, java_ref):
        self._java_ref = java_ref

    def __repr__(self):
        return "pygw {} => {}".format(self.__class__, self._java_ref)

    def __eq__(self, other):
        if not isinstance(other, GeoWaveObject):
            return False
        return self._java_ref == other._java_ref

    def __str__(self):
        return self.to_string()

    def is_instance_of(self, java_class):
        """
        Returns:
            True if this object is of the type represented by the given java class.
        """
        return is_instance_of(java_gateway, self._java_ref, java_class)

    def to_string(self):
        return self._java_ref.toString()

    def java_ref(self):
        return self._java_ref

    @staticmethod
    def to_java_array(java_class, objects):
        n = len(objects)
        j_arr = java_gateway.new_array(java_class, n)
        for idx, obj in enumerate(objects):
            if not isinstance(obj, GeoWaveObject) or not obj.is_instance_of(java_class):
                print(obj, objects, java_class)
                raise AttributeError("Given object is not compatible with the given class.")
            j_arr[idx] = obj.java_ref()

        return j_arr



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.base.java_transformer import JavaTransformer
from pygw.base.type_conversions import PrimitiveByteArrayType


class BinnedStatisticTransformer(JavaTransformer):
    """
    Transforms a binned statistic value into a tuple.
    """

    def __init__(self, statistic_value_transformer):
        self._statistic_value_transformer = statistic_value_transformer
        self._byte_array_type = PrimitiveByteArrayType()
        super().__init__()

    def transform(self, j_object):
        """
        Transform the given bin/value pair into a tuple.

        Args:
            j_object (Java Pair): The bin/value pair.
        Returns:
            A tuple of bin bytes and statistic value.
        """
        return self._byte_array_type.from_java(j_object.getKey().getBytes()),\
            self._statistic_value_transformer.transform(j_object.getValue())



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.base import GeoWaveObject
from pygw.config import geowave_pkg
from ..statistic_binning_strategy import StatisticBinningStrategy


class CompositeBinningStrategy(StatisticBinningStrategy):
    """
    Statistic binning strategy that combines two or more other binning strategies.
    """

    def __init__(self, binning_strategies=None, java_ref=None):
        if java_ref is None:
            j_binning_strategies = GeoWaveObject.to_java_array(
                geowave_pkg.core.store.api.StatisticBinningStrategy, binning_strategies)
            java_ref = geowave_pkg.core.store.statistics.binning.CompositeBinningStrategy(j_binning_strategies)
        super().__init__(java_ref)
